# SPDX-License-Identifier: Apache-2.0

from .main import load_model
from .main import save_model
from .main import set_model_version
from .main import set_model_domain
from .main import set_model_doc_string
from onnxconverter_common.metadata_props import add_metadata_props, set_denotation
from .visualize import visualize_model
from .float16_converter import convert_float_to_float16
from .tests_helper import dump_data_and_model
from .tests_helper import dump_one_class_classification, dump_binary_classification, dump_multiple_classification
from .tests_helper import dump_multiple_regression, dump_single_regression
from .tests_dl_helper import create_tensor



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ...common._registration import register_converter


def convert_glm_regressor(scope, operator, container):
    from coremltools.proto.GLMRegressor_pb2 import GLMRegressor

    op_type = 'LinearRegressor'
    glm = operator.raw_operator.glmRegressor
    attrs = {'name': operator.full_name}

    transform_table = {GLMRegressor.NoTransform: 'NONE', GLMRegressor.Logit: 'LOGISTIC', GLMRegressor.Probit: 'PROBIT'}
    if glm.postEvaluationTransform in transform_table:
        attrs['post_transform'] = transform_table[glm.postEvaluationTransform]
    else:
        raise ValueError('Unsupported post-transformation: {}'.format(glm.postEvaluationTransform))

    # Determine the dimensionality of the model weights. Conceptually,
    # the shape of the weight matrix in CoreML is E-by-F, where E and F
    # respectively denote the number of target variables and the number
    # of used features. Note that in ONNX, the shape is F-by-E.
    dim_target = len(glm.weights)
    dim_feature = len(glm.weights[0].value)

    matrix_w = np.ndarray(shape=(dim_feature, dim_target))
    for i, w in enumerate(glm.weights):
        matrix_w[:, i] = w.value

    attrs['targets'] = dim_target
    attrs['coefficients'] = matrix_w.flatten()
    attrs['intercepts'] = glm.offset

    container.add_node(op_type, operator.input_full_names, operator.output_full_names, op_domain='ai.onnx.ml', **attrs)


register_converter('glmRegressor', convert_glm_regressor)



# SPDX-License-Identifier: Apache-2.0

from ...common.data_types import Int64TensorType, Int64Type, FloatTensorType, FloatType, StringType


def convert_integer_to_float(scope, variable, container):
    op_type = 'Scaler'
    scaled_name = scope.get_unique_variable_name(variable.full_name + '_scaled')
    scaler_attrs = {'name': scope.get_unique_operator_name(op_type), 'scale': [1.], 'offset': [0.]}
    container.add_node('Scaler', variable.full_name, scaled_name, op_domain='ai.onnx.ml', **scaler_attrs)
    return scaled_name


def concatenate_variables(scope, variables, container):
    '''
    This function allocate operators to from a float tensor by concatenating all input variables. Notice that if all
    integer inputs would be converted to floats before concatenation.
    '''

    # Check if it's possible to concatenate those inputs.
    type_set = set(type(variable.type) for variable in variables)
    number_type_set = {FloatType, FloatTensorType, Int64Type, Int64TensorType}
    if StringType in type_set and any(number_type in type_set for number_type in number_type_set):
        raise RuntimeError('We are not able to concatenate numerical tensor(s) and string tensor(s)')

    input_names = []  # input variables' names we want to concatenate
    input_dims = []  # dimensions of the variables that is going to be concatenated

    # Collect input variable names and do cast if needed
    for variable in variables:
        if isinstance(variable.type, (Int64TensorType, Int64Type)):
            input_names.append(convert_integer_to_float(scope, variable, container))
        else:
            input_names.append(variable.full_name)
        # We assume input variables' shape are [1, C_1], ..., [1, C_n] if there are n inputs.
        input_dims.append(variable.type.shape[1])

    if len(input_names) == 1:
        # No need to concatenate tensors if there is only one input
        return input_names[0]
    else:
        # To combine all inputs, we need a FeatureVectorizer
        op_type = 'FeatureVectorizer'
        attrs = {'name': scope.get_unique_operator_name(op_type), 'inputdimensions': input_dims}
        # Create a variable name to capture feature vectorizer's output
        concatenated_name = scope.get_unique_variable_name('concatenated')
        # Set up our FeatureVectorizer
        container.add_node(op_type, input_names, concatenated_name, op_domain='ai.onnx.ml', **attrs)

        return concatenated_name



# SPDX-License-Identifier: Apache-2.0

from ....proto import helper
from ....proto import onnx_proto
from ...common._registration import register_converter
from ...common._apply_operation import apply_constant

def convert_tensor_to_label(scope, operator, container):
    '''
    This converter tries to convert a dummy operator 'TensorToLabel' into a sequence of some ONNX operators. Those
    operators are used to extract the label with the highest probability for doing a prediction. We assume that the
    elements in the given probability tensor are aligned with the class labels specified in the CoreML model. That is,
    if you have a class label vector ['a', 'b'] in our CoreML classifier, the first (and the only) input of this
    operator should be [probability_of_class_a, probability_of_class_b].

    Assume that we have C classes with batch size N (N must be 1. If not, the output class probabilities need to be
    encoded as a sequence of dictionary, which is not allowed in ONNX). The ONNX computation graph of this operator may
    look like

                Probability tensor [1, C] (the variable defined at operator.inputs[0])
                        |
                        v
                      ArgMax                                        LoadConstant (its attribute is extracted from
                        |                                                |        operator.raw_operator, which is a
                        |                                                |        CoreML classifier)
                        |                                                |
                        v                                                |
                   best index [1]                                        |
                        |                                                |
                        v                                                v
               ArrayFeatureExtractor  <-------------------- a 1-D tensor of class labels [C]
                        |
                        v
                  predicted label [1]
    '''
    model_type = operator.raw_operator.WhichOneof('Type')
    if model_type == 'neuralNetworkClassifier':
        model = operator.raw_operator.neuralNetworkClassifier
        if model.WhichOneof('ClassLabels') == 'stringClassLabels':
            labels = list(s.encode('utf-8') for s in model.stringClassLabels.vector)
            label_type = onnx_proto.TensorProto.STRING
        elif model.WhichOneof('ClassLabels') == 'int64ClassLabels':
            labels = list(int(i) for i in model.int64ClassLabels.vector)
            label_type = onnx_proto.TensorProto.INT64
        else:
            raise ValueError('Unknown label type found')
    elif model_type == 'pipelineClassifier':
        model = operator.raw_operator.pipelineClassifier
        if model.WhichOneof('ClassLabels') == 'stringClassLabels':
            labels = list(s.encode('utf-8') for s in model.pipelineClassifier.stringClassLabels.vector)
            label_type = onnx_proto.TensorProto.STRING
        elif model.WhichOneof('ClassLabels') == 'int64ClassLabels':
            labels = list(int(i) for i in model.int64ClassLabels.vector)
            label_type = onnx_proto.TensorProto.INT64
        else:
            raise ValueError('Unknown label type found')
    else:
        raise ValueError('Only neural network classifiers and pipeline classifiers are supported')

    # Use a Constant operator to load and output all labels as a tensor
    label_loader_name = scope.get_unique_operator_name('LabelLoader')
    label_buffer_name = scope.get_unique_variable_name('ClassLabels')
    label_loader_value = helper.make_tensor(label_buffer_name, label_type, [len(labels)], labels)
    apply_constant(scope, [label_buffer_name], container,
                    operator_name=label_loader_name, value=label_loader_value)

    # Extract most possible label index
    label_id_extractor_name = scope.get_unique_operator_name('LabelIndexExtractor')
    label_id_extractor_attrs = {'name': label_id_extractor_name}
    label_id_extractor_attrs['axis'] = 1
    label_id_extractor_attrs['keepdims'] = 1
    extracted_id_name = scope.get_unique_variable_name('LabelId')
    container.add_node('ArgMax', [operator.inputs[0].full_name], [extracted_id_name], **label_id_extractor_attrs)

    # Pick up the label indicated by the selected ID
    label_selector_name = scope.get_unique_operator_name('LabelSelector')
    label_selector_attrs = {'name': label_selector_name}
    container.add_node('ArrayFeatureExtractor', [label_buffer_name, extracted_id_name], [operator.outputs[0].full_name],
                       op_domain='ai.onnx.ml', **label_selector_attrs)


register_converter('tensorToLabel', convert_tensor_to_label)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark.ml.feature import PolynomialExpansion
from pyspark.ml.linalg import Vectors
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlPolynomialExpansion(SparkMlTestCase):

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_model_polynomial_expansion(self):
        data = self.spark.createDataFrame([
            (Vectors.dense([1.2, 3.2, 1.3, -5.6]),),
            (Vectors.dense([4.3, -3.2, 5.7, 1.0]),),
            (Vectors.dense([0, 3.2, 4.7, -8.9]),)
        ], ["dense"])
        model = PolynomialExpansion(degree=2, inputCol="dense", outputCol="expanded")

        # the input name should match that of what StringIndexer.inputCol
        feature_count = data.first()[0].size
        model_onnx = convert_sparkml(model, 'Sparkml PolynomialExpansion', [('dense', FloatTensorType([None, feature_count]))],
                                     target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        expected = predicted.toPandas().expanded.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        data_np = data.toPandas().dense.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlPolynomialExpansion")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['expanded'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import numpy as np


class Node:

    _names_classifier = [
        "class_ids",
        "class_nodeids",
        "class_treeids",
        "class_weights",
        "classlabels_int64s",
        "nodes_falsenodeids",
        "nodes_featureids",
        "nodes_hitrates",
        "nodes_missing_value_tracks_true",
        "nodes_modes",
        "nodes_nodeids",
        "nodes_treeids",
        "nodes_truenodeids",
        "nodes_values",
    ]

    _names_regressor = [
        "target_ids",
        "target_nodeids",
        "target_treeids",
        "target_weights",
        "nodes_falsenodeids",
        "nodes_featureids",
        "nodes_hitrates",
        "nodes_missing_value_tracks_true",
        "nodes_modes",
        "nodes_nodeids",
        "nodes_treeids",
        "nodes_truenodeids",
        "nodes_values",
    ]

    def __init__(self, is_classifier, **kwargs):
        if is_classifier:
            self.is_classifier = True
            self._names = Node._names_classifier
        else:
            self.is_classifier = False
            self._names = Node._names_regressor
        for att in self._names:
            setattr(self, att, None)

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        ps = ", ".join(map(lambda k: f"{k}={getattr(self, k)!r}", self._names))
        return f"Node({ps})"

    @property
    def attrs(self):
        return {k: getattr(self, k) for k in self._names}

    @staticmethod
    def create(attrs):
        nodes = {}
        root = None
        indices = attrs["nodes_nodeids"]
        for n, nid in enumerate(indices):
            tid = attrs["nodes_treeids"][n]
            mode = attrs["nodes_modes"][n]
            kwargs = {}
            for k, v in attrs.items():
                if not k.startswith("nodes"):
                    continue
                kwargs[k] = v[n]
            if mode == "LEAF":
                if "class_treeids" in attrs:
                    # classifier
                    pos = [
                        i
                        for i, (t, c) in enumerate(
                            zip(attrs["class_treeids"], attrs["class_nodeids"])
                        )
                        if t == tid and c == nid
                    ]
                else:
                    # regressor
                    pos = [
                        i
                        for i, (t, c) in enumerate(
                            zip(attrs["target_treeids"], attrs["target_nodeids"])
                        )
                        if t == tid and c == nid
                    ]                    
                for k, v in attrs.items():
                    if k in {"post_transform", "name", "domain", "n_targets"}:
                        continue
                    if k.startswith("nodes"):
                        continue
                    if "label" in k:
                        kwargs[k] = v
                        continue
                    try:
                        kwargs[k] = [v[p] for p in pos]
                    except TypeError as e:
                        raise TypeError(f"Unabel to update attribute {k!r}.") from e

            node = Node("class_treeids" in attrs, **kwargs)
            if mode == "BRANCH_LEQ" and isinstance(
                node.nodes_values, (np.ndarray, list)
            ):
                node.nodes_modes = "||"
            nodes[tid, nid] = node
            if root is None:
                root = node

        # add links
        for k, v in nodes.items():
            if v.nodes_modes != "LEAF":
                ntrue = nodes[v.nodes_treeids, v.nodes_truenodeids]
                nfalse = nodes[v.nodes_treeids, v.nodes_falsenodeids]
                v._nodes_truenode = ntrue
                v._nodes_falsenode = nfalse
        return root, nodes

    def enumerate_nodes(self, unique=False):
        if unique:
            done = set()
            yield self
            done.add(id(self))
            if self.nodes_modes != "LEAF":
                for node in self._nodes_truenode:
                    if id(node) not in done:
                        yield node
                        done.add(id(node))
                for node in self._nodes_falsenode:
                    if id(node) not in done:
                        yield node
                        done.add(id(node))
        else:
            yield self
            if self.nodes_modes != "LEAF":
                for node in self._nodes_truenode:
                    yield node
                for node in self._nodes_falsenode:
                    yield node

    def __iter__(self):
        "Iterates over the node."
        for node in self.enumerate_nodes(unique=True):
            yield node

    def _unfold_rule_or(self):
        if self.nodes_modes == "||":
            values = self.nodes_values
            if len(values) == 1:
                self.nodes_modes = "BRANCH_EQ"
                self.nodes_values = self.nodes_values[0]
                return self
            if len(values) == 0:
                raise ValueError(f"Issue with {self!r}.")
            th = self.nodes_values[-1]
            vals = self.nodes_values[:-1]

            new_node = Node(self.is_classifier, **self.attrs)
            new_node.nodes_values = th
            new_node.nodes_modes = "BRANCH_EQ"
            new_node._nodes_truenode = self._nodes_truenode
            new_node._nodes_falsenode = self._nodes_falsenode

            self.nodes_values = vals
            self._nodes_falsenode = new_node
            return self
        return None

    def unfold_rule_or(self):
        cond = True
        nodes = []
        for node in self:
            if node.nodes_modes == "||":
                nodes.append(node)
        if len(nodes) == 0:
            return
        for node in nodes:
            r = node._unfold_rule_or()
            if r is not None:
                while r is not None:
                    r = r._unfold_rule_or()
        self.set_new_numbers()

    def set_new_numbers(self):
        # new number
        tree = {}
        done = set()
        stack = [self]
        while len(stack) > 0:
            node = stack[0]
            stack = stack[1:]
            if id(node) in done:
                continue
            done.add(id(node))
            tid = node.nodes_treeids
            if tid not in tree:
                tree[tid] = -1
            nid = tree[tid] + 1
            tree[tid] += 1
            node.nodes_nodeids = nid
            if node.nodes_modes != "LEAF":
                stack.append(node._nodes_truenode)
                stack.append(node._nodes_falsenode)
        # all ids
        ids = {}
        for node in self:
            ids[id(node)] = node.nodes_nodeids
        # replaces true, false ids
        for node in self:
            if node.nodes_modes == "LEAF":
                continue
            node.nodes_falsenodeids = ids[id(node._nodes_falsenode)]
            node.nodes_truenodeids = ids[id(node._nodes_truenode)]

    def to_attrs(self, **kwargs):
        """
        Returns the nodes as node attributes for ONNX.
        """
        nodes = {}
        for node in self:
            nodes[node.nodes_treeids, node.nodes_nodeids] = node
        sort = [v for k, v in sorted(nodes.items())]
        attrs = {}
        for name in self._names:
            attrs[name] = []
        for node in sort:
            for k in self._names:
                if not k.startswith("nodes"):
                    continue
                attrs[k].append(getattr(node, k))
            if node.nodes_modes == "LEAF":
                for k in self._names:
                    if k.startswith("nodes"):
                        continue
                    if k not in attrs:
                        attrs[k] = []
                    if k == "class_nodeids":
                        attrs[k].extend([node.nodes_nodeids for _ in node.class_ids])
                    else:
                        try:
                            attrs[k].extend(getattr(node, k))
                        except TypeError as e:
                            raise TypeError(f"Issue with attribute {k!r}.") from e
        attrs.update(kwargs)

        # update numbers
        new_numbers = {}
        for tid, nid, md in sorted(zip(attrs["nodes_treeids"], attrs["nodes_nodeids"], attrs["nodes_modes"])):
            new_numbers[tid, nid] = len(new_numbers)
        for k in ["nodes_truenodeids", "nodes_falsenodeids", "nodes_nodeids", "class_nodeids", "target_nodeids"]:
            if k not in attrs:
                continue
            if "class_" in k or "target_" in k:
                field = k.split('_')[0] + "_treeids"
            else:
                field = "nodes_treeids"
            for i in range(len(attrs[k])):
                nid = attrs[k][i]
                if nid == 0 and k in {'nodes_truenodeids', 'nodes_falsenodeids'}:
                    continue
                tid = attrs[field][i]
                new_id = new_numbers[tid, nid]
                attrs[k][i] = new_id
        return attrs


def rewrite_ids_and_process(attrs, logger):
    in_sets_rules = []
    for i, value in enumerate(attrs["nodes_values"]):
        if isinstance(value, (np.ndarray, list)):
            in_sets_rules.append(i)

    logger.info("[convert_decision_tree_classifier] in_set_rules has %d elements", len(in_sets_rules))
    for i in in_sets_rules:
        attrs["nodes_modes"][i] = "||"
    logger.info("[convert_decision_tree_classifier] Node.create")
    root, _ = Node.create(attrs)
    logger.info("[convert_decision_tree_classifier] unfold_rule_or")
    root.unfold_rule_or()
    logger.info("[convert_decision_tree_classifier] to_attrs")
    if "class_nodeids" in attrs:
        new_attrs = root.to_attrs(
                post_transform=attrs['post_transform'],
                classlabels_int64s=attrs["classlabels_int64s"],
                name=attrs['name'])
    else:
        new_attrs = root.to_attrs(
                post_transform=attrs['post_transform'],
                n_targets=attrs['n_targets'],
                name=attrs['name'])
    if len(attrs['nodes_nodeids']) > len(new_attrs['nodes_nodeids']):
        raise RuntimeError(
            f"The replacement fails as there are less nodes in the new tree, "
            f"{len(attrs['nodes_nodeids'])} > {len(new_attrs['nodes_nodeids'])}."
        )
    if set(attrs) != set(new_attrs):
        raise RuntimeError(f"Missing key: {list(sorted(attrs))} != {list(sorted(new_attrs))}.")
    logger.info("[convert_decision_tree_classifier] n_nodes=%d", len(attrs['nodes_nodeids']))
    logger.info("[convert_decision_tree_classifier] end")
    return new_attrs



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter
from ...common._registration import register_shape_calculator
from ...common.utils import check_input_and_output_numbers
from ...common.data_types import *


def convert_sparkml_vector_assembler(scope, operator, container):
    container.add_node('Concat', [s for s in operator.input_full_names], operator.outputs[0].full_name,
                       name=scope.get_unique_operator_name('Concat'),
                       op_version=4,
                       axis=1)


register_converter('pyspark.ml.feature.VectorAssembler', convert_sparkml_vector_assembler)


def calculate_vector_assembler_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)

    # Sum up the rank 1 (length of input vector) from each input.
    C = sum([input.type.shape[1] for input in operator.inputs])
    N = operator.inputs[0].type.shape[0]
    col_type = operator.inputs[0].type
    if isinstance(col_type, FloatTensorType):
        col_type = FloatTensorType([N, C])
    elif isinstance(col_type, Int64TensorType):
        col_type = Int64TensorType([N, C])
    else:
        raise TypeError("Unsupported input type")
    operator.outputs[0].type = col_type


register_shape_calculator('pyspark.ml.feature.VectorAssembler', calculate_vector_assembler_shapes)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark.ml.feature import Word2Vec
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import StringTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


## For some reason during the spark bring up and shutdown something happens causing these
## tests to fail. For that you need to run each test here individually
## For now these will be commented out so as not to break the build
##      AttributeError: 'NoneType' object has no attribute 'setCallSite' on model.surrogateDF
##  Therefore we leave these tests out for now until a newere version of pyspark is availabe that address this issue
class TestSparkmlWord2Vec(SparkMlTestCase):

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_word2vec(self):
        data = self.spark.createDataFrame([
            ("Hi I heard about Spark".split(" "), ),
            ("I wish Java could use case classes".split(" "), ),
            ("Logistic regression models are neat".split(" "), )
        ], ["text"])
        word2Vec = Word2Vec(vectorSize=3, minCount=0, inputCol="text", outputCol="result")
        model = word2Vec.fit(data)
        vectors = model.getVectors()
        vectors.show(100, False)
    
        result = model.transform(data)
        result.show(100, False)
    
        # the input name should match that of inputCol
        feature_count = len(data.first()[0])
        model_onnx = convert_sparkml(model, 'Sparkml Word2Vec', [('text', StringTensorType([None, feature_count]))],
                                     target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data.limit(1))
        expected = predicted.toPandas().result.apply(lambda  x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        data_np = data.limit(1).toPandas().text.values
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlWord2Vec")
        onnx_model_path = paths[-1]
        data_np = numpy.array(data_np[0]).reshape((1, -1))
        output, output_shapes = run_onnx_model(['result'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import warnings
import packaging.version as pv
import onnx
from .common import utils


def convert_coreml(model, name=None, initial_types=None, doc_string='', target_opset=None,
                   targeted_onnx=None, custom_conversion_functions=None, custom_shape_calculators=None):
    if targeted_onnx is not None:
        warnings.warn("targeted_onnx is deprecated. Use target_opset.", DeprecationWarning)
    if not utils.coreml_installed():
        raise RuntimeError('coremltools is not installed. Please install coremltools to use this feature.')

    from .coreml.convert import convert
    return convert(model, name, initial_types, doc_string, target_opset, targeted_onnx,
                   custom_conversion_functions, custom_shape_calculators)


def convert_keras(model, name=None,
                  initial_types=None,
                  doc_string='',
                  target_opset=None,
                  targeted_onnx=None,
                  channel_first_inputs=None,
                  custom_conversion_functions=None,
                  custom_shape_calculators=None,
                  default_batch_size=1):
    """
    .. versionchanged:: 1.9.0
        The conversion is now using *tf2onnx*.
    """
    if targeted_onnx is not None:
        warnings.warn("targeted_onnx is deprecated and unused. Use target_opset.", DeprecationWarning)
    import tensorflow as tf
    if pv.Version(tf.__version__) < pv.Version('2.0'):
        # Former converter for tensorflow<2.0.
        from keras2onnx import convert_keras as convert
        return convert(model, name, doc_string, target_opset, channel_first_inputs)
    else:
        # For tensorflow>=2.0, new converter based on tf2onnx.
        import tf2onnx

        if not utils.tf2onnx_installed():
            raise RuntimeError('tf2onnx is not installed. Please install it to use this feature.')

        if custom_conversion_functions is not None:
            warnings.warn('custom_conversion_functions is not supported any more. Please set it to None.')
        if custom_shape_calculators is not None:
            warnings.warn('custom_shape_calculators is not supported any more. Please set it to None.')
        if default_batch_size != 1:
            warnings.warn('default_batch_size is not supported any more. Please set it to 1.')
        if default_batch_size != 1:
            warnings.warn('default_batch_size is not supported any more. Please set it to 1.')

        if initial_types is not None:
            from onnxconverter_common import (
                FloatTensorType, DoubleTensorType,
                Int64TensorType, Int32TensorType,
                StringTensorType, BooleanTensorType)
            spec = []
            for name, kind in initial_types:
                if isinstance(kind, FloatTensorType):
                    dtype = tf.float32
                elif isinstance(kind, Int64TensorType):
                    dtype = tf.int64
                elif isinstance(kind, Int32TensorType):
                    dtype = tf.int32
                elif isinstance(kind, DoubleTensorType):
                    dtype = tf.float64
                elif isinstance(kind, StringTensorType):
                    dtype = tf.string
                elif isinstance(kind, BooleanTensorType):
                    dtype = tf.bool
                else:
                    raise TypeError(
                        "Unexpected type %r, cannot infer tensorflow type." % type(kind))
                spec.append(tf.TensorSpec(tuple(kind.shape), dtype, name=name))
            input_signature = tuple(spec)
        else:
            input_signature = None

        model_proto, external_tensor_storage = tf2onnx.convert.from_keras(
            model,
            input_signature=input_signature,
            opset=target_opset,
            custom_ops=None,
            custom_op_handlers=None,
            custom_rewriter=None,
            inputs_as_nchw=channel_first_inputs,
            extra_opset=None,
            shape_override=None,
            target=None,
            large_model=False,
            output_path=None)
        if external_tensor_storage is not None:
            warnings.warn("The current API does not expose the second result 'external_tensor_storage'. "
                          "Use tf2onnx directly to get it.")
        model_proto.doc_string = doc_string
        return model_proto


def convert_libsvm(model, name=None, initial_types=None, doc_string='', target_opset=None,
                   targeted_onnx=None, custom_conversion_functions=None, custom_shape_calculators=None):
    if targeted_onnx is not None:
        warnings.warn("targeted_onnx is deprecated. Use target_opset.", DeprecationWarning)
    if not utils.libsvm_installed():
        raise RuntimeError('libsvm is not installed. Please install libsvm to use this feature.')

    from .libsvm.convert import convert
    return convert(model, name, initial_types, doc_string, target_opset, targeted_onnx,
                   custom_conversion_functions, custom_shape_calculators)


def convert_catboost(model, name=None, initial_types=None, doc_string='', target_opset=None):
    try:
        from catboost.utils import convert_to_onnx_object
    except ImportError:
        raise RuntimeError('CatBoost is not installed or needs to be updated. '
                           'Please install/upgrade CatBoost to use this feature.')

    return convert_to_onnx_object(model, export_parameters={'onnx_doc_string': doc_string, 'onnx_graph_name': name},
                                  initial_types=initial_types, target_opset=target_opset)


def convert_lightgbm(model, name=None, initial_types=None, doc_string='', target_opset=None,
                     targeted_onnx=None, custom_conversion_functions=None,
                     custom_shape_calculators=None, without_onnx_ml=False, zipmap=True,
                     split=None):
    if targeted_onnx is not None:
        warnings.warn("targeted_onnx is deprecated. Use target_opset.", DeprecationWarning)
    if not utils.lightgbm_installed():
        raise RuntimeError('lightgbm is not installed. Please install lightgbm to use this feature.')

    from .lightgbm.convert import convert
    return convert(model, name, initial_types, doc_string, target_opset, targeted_onnx,
                   custom_conversion_functions, custom_shape_calculators, without_onnx_ml,
                   zipmap=zipmap, split=split)


def convert_sklearn(model, name=None, initial_types=None, doc_string='', target_opset=None,
                    targeted_onnx=None, custom_conversion_functions=None, custom_shape_calculators=None):
    if targeted_onnx is not None:
        warnings.warn("targeted_onnx is deprecated. Use target_opset.", DeprecationWarning)
    if not utils.sklearn_installed():
        raise RuntimeError('scikit-learn is not installed. Please install scikit-learn to use this feature.')

    if not utils.skl2onnx_installed():
        raise RuntimeError('skl2onnx is not installed. Please install skl2onnx to use this feature.')

    from skl2onnx.convert import convert_sklearn as convert_skl2onnx
    return convert_skl2onnx(model, name, initial_types, doc_string, target_opset,
                            custom_conversion_functions, custom_shape_calculators)


def convert_sparkml(model, name=None, initial_types=None, doc_string='', target_opset=None,
                    targeted_onnx=None, custom_conversion_functions=None,
                    custom_shape_calculators=None, spark_session=None):
    if targeted_onnx is not None:
        warnings.warn("targeted_onnx is deprecated. Use target_opset.", DeprecationWarning)
    if not utils.sparkml_installed():
        raise RuntimeError('Spark is not installed. Please install Spark to use this feature.')

    from .sparkml.convert import convert
    return convert(model, name, initial_types, doc_string, target_opset, targeted_onnx,
                   custom_conversion_functions, custom_shape_calculators, spark_session)


def convert_xgboost(*args, **kwargs):
    if kwargs.get('targeted_onnx', None) is not None:
        warnings.warn("targeted_onnx is deprecated. Use target_opset.", DeprecationWarning)
    if not utils.xgboost_installed():
        raise RuntimeError('xgboost is not installed. Please install xgboost to use this feature.')

    from .xgboost.convert import convert
    return convert(*args, **kwargs)


def convert_h2o(*args, **kwargs):
    if kwargs.get('targeted_onnx', None) is not None:
        warnings.warn("targeted_onnx is deprecated. Use target_opset.", DeprecationWarning)
    if not utils.h2o_installed():
        raise RuntimeError('h2o is not installed. Please install h2o to use this feature.')

    from .h2o.convert import convert
    return convert(*args, **kwargs)


def _collect_input_nodes(graph, outputs):
    nodes_to_keep = set()
    input_nodes = set()
    node_inputs = [graph.get_tensor_by_name(ts_).op for ts_ in outputs]
    while node_inputs:
        nd_ = node_inputs[0]
        del node_inputs[0]
        if nd_.type in ['Placeholder', "PlaceholderV2", 'PlaceholderWithDefault']:
            input_nodes.add(nd_)
        if nd_ in nodes_to_keep:
            continue

        nodes_to_keep.add(nd_)
        node_inputs.extend(in_.op for in_ in nd_.inputs)

    return input_nodes, nodes_to_keep


def _convert_tf_wrapper(frozen_graph_def,
                        name=None, input_names=None, output_names=None,
                        doc_string='',
                        target_opset=None,
                        channel_first_inputs=None,
                        debug_mode=False, custom_op_conversions=None,
                        **kwargs):
    """
    convert a tensorflow graph def into a ONNX model proto, just like how keras does.
    :param graph_def: the frozen tensorflow graph
    :param name: the converted onnx model internal name
    :param input_names: the inputs name list of the model
    :param output_names: the output name list of the model
    :param doc_string: doc string
    :param target_opset: the targeted onnx model opset
    :param channel_first_inputs: A list of channel first input (not supported yet)
    :param debug_mode: will enable the log and try to convert as much as possible on conversion
    :param kwargs: additional parameters of function `processs_tf_graph
        <https://github.com/onnx/tensorflow-onnx#creating-custom-op-mappings-from-python>`_
    :return an ONNX ModelProto
    """
    import tensorflow as tf
    import tf2onnx

    if target_opset is None:
        target_opset = onnx.defs.onnx_opset_version()

    if not doc_string:
        doc_string = "converted from {}".format(name)

    tf_graph_def = tf2onnx.tfonnx.tf_optimize(input_names, output_names, frozen_graph_def, True)
    with tf.Graph().as_default() as tf_graph:
        tf.import_graph_def(tf_graph_def, name='')

        if not input_names:
            input_nodes = list(_collect_input_nodes(tf_graph, output_names)[0])
            input_names = [nd_.outputs[0].name for nd_ in input_nodes]
        g = tf2onnx.tfonnx.process_tf_graph(tf_graph,
                                            continue_on_error=debug_mode,
                                            opset=target_opset,
                                            custom_op_handlers=custom_op_conversions,
                                            inputs_as_nchw=channel_first_inputs,
                                            output_names=output_names,
                                            input_names=input_names,
                                            **kwargs)

        onnx_graph = tf2onnx.optimizer.optimize_graph(g)
        model_proto = onnx_graph.make_model(doc_string)

    return model_proto


def convert_tensorflow(frozen_graph_def,
                       name=None, input_names=None, output_names=None,
                       doc_string='',
                       target_opset=None,
                       channel_first_inputs=None,
                       debug_mode=False, custom_op_conversions=None,
                       **kwargs):
    import pkgutil
    if not pkgutil.find_loader('tf2onnx'):
        raise RuntimeError('tf2onnx is not installed, please install it before calling this function.')

    return _convert_tf_wrapper(frozen_graph_def, name, input_names, output_names, doc_string,
                               target_opset, channel_first_inputs, debug_mode, custom_op_conversions,
                               **kwargs)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_normalizer(scope, operator, container):
    op_type = 'Normalizer'
    attrs = {'name': operator.full_name}
    norms = ['MAX', 'L1', 'L2']
    norm_type = operator.raw_operator.normalizer.normType
    if norm_type in range(3):
        attrs['norm'] = norms[norm_type]
    else:
        raise ValueError('Invalid norm type: ' + norm_type)

    container.add_node(op_type, operator.input_full_names, operator.output_full_names, op_domain='ai.onnx.ml', **attrs)


register_converter('normalizer', convert_normalizer)



# SPDX-License-Identifier: Apache-2.0

import sys
import inspect
import unittest
import packaging.version as pv
import os
import onnx
import pandas
import numpy
from pyspark.ml.linalg import VectorUDT, SparseVector
from pyspark.ml.regression import RandomForestRegressor
from pyspark.ml import Pipeline
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType, StringTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase
from pyspark.ml.feature import VectorIndexer, StringIndexer


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmRandomForestRegressor(SparkMlTestCase):

    @unittest.skipIf(sys.platform == 'win32',
                     reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    @unittest.skipIf(pv.Version(onnx.__version__) <= pv.Version('1.3'),
                     'Need Greater Opset 9')
    def test_random_forest_regression(self):
        this_script_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        input_path = os.path.join(this_script_dir, "data", "sample_libsvm_data.txt")
        original_data = self.spark.read.format("libsvm").load(input_path)
        #
        # truncate the features
        #
        feature_count = 5
        self.spark.udf.register("truncateFeatures",
                                lambda x: SparseVector(feature_count, range(0,feature_count), x.toArray()[125:130]),
                                VectorUDT())
        data = original_data.selectExpr("cast(label as string) as label", "truncateFeatures(features) as features")
        label_indexer = StringIndexer(inputCol="label", outputCol="indexedLabel")
        feature_indexer = VectorIndexer(inputCol="features", outputCol="indexedFeatures",
                                        maxCategories=10, handleInvalid='error')

        rf = RandomForestRegressor(labelCol="indexedLabel", featuresCol="indexedFeatures", numTrees=3)
        pipeline = Pipeline(stages=[label_indexer, feature_indexer, rf])
        model = pipeline.fit(data)
        model_onnx = convert_sparkml(model, 'Sparkml RandomForest Regressor', [
            ('label', StringTensorType([None, 1])),
            ('features', FloatTensorType([None, feature_count]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data.limit(1))
        data_np = {
            'label': data.limit(1).toPandas().label.values.reshape((-1, 1)),
            'features': data.limit(1).toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        }
        expected = [
            predicted.toPandas().indexedLabel.values.astype(numpy.int64),
            predicted.toPandas().prediction.values.astype(numpy.float32)
        ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlRandomForestRegressor")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['indexedLabel', 'prediction'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter
from .....proto import onnx_proto


def convert_slice(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import SliceLayerParams as Params

    op_type = 'Slice'
    op_name = scope.get_unique_operator_name(op_type)
    attrs = {'name': op_name}
    params = operator.raw_operator.slice

    # Set up slice range of C-, H-, and W-axes. Notice that only one of them will be actually sliced.
    axis_map = {Params.CHANNEL_AXIS: 0, Params.HEIGHT_AXIS: 1, Params.WIDTH_AXIS: 2}
    starts = [0, 0, 0]
    ends = [-1, -1, -1]
    starts[axis_map[params.axis]] = params.startIndex
    ends[axis_map[params.axis]] = params.endIndex

    if params.stride != 1:
        raise ValueError('Stride must be 1 but got %s' % params.stride)

    if container.target_opset < 10:
        # The input shape should be [N, C, H, W] in ONNX. Because CoreML only slices one of C-, H-, or W-axes, the
        # "axes" attribute in ONNX is [1, 2, 3]. Note that for the axes not really sliced, their starting and ending
        # indexes are 0 and -1, respectively.
        attrs['axes'] = [1, 2, 3]
        attrs['starts'] = starts
        attrs['ends'] = ends
        op_version = 1
        container.add_node(op_type, operator.input_full_names, operator.output_full_names, op_version=op_version, **attrs)
    else:
        starts_name = scope.get_unique_variable_name(operator.full_name + '_starts')
        container.add_initializer(starts_name, onnx_proto.TensorProto.INT64,
                                  [3], starts)
        ends_name = scope.get_unique_variable_name(operator.full_name + '_ends')
        container.add_initializer(ends_name, onnx_proto.TensorProto.INT64,
                                  [3], ends)
        axes_name = scope.get_unique_variable_name(operator.full_name + '_axes')
        container.add_initializer(axes_name, onnx_proto.TensorProto.INT64,
                                  [3], [1, 2, 3])

        if container.target_opset < 11:
            op_version = 10
        else:
            op_version = 11
        container.add_node(op_type, [operator.input_full_names[0], starts_name, ends_name, axes_name], operator.output_full_names, op_version=op_version,
                           **attrs)


register_converter('slice', convert_slice)



# SPDX-License-Identifier: Apache-2.0

import itertools
import numpy as np
from ...common._registration import register_converter


def extract_support_vectors_as_dense_tensor(svm_model):
    support_type = svm_model.WhichOneof('supportVectors')
    if support_type == 'denseSupportVectors':
        vectors = svm_model.denseSupportVectors.vectors
        support_vectors = np.array([v.values for v in vectors]).flatten()
    elif support_type == 'sparseSupportVectors':
        # Since current ONNX doesn't support sparse representations, we have to save them as dense tensors. It may
        # dramatically prolong prediction time and increase memory usage.
        vectors = svm_model.sparseSupportVectors.vectors
        # Search for the maximum dimension over all sparse support vectors
        max_idx = 0
        for v in vectors:
            for n in v.nodes:
                if n.index > max_idx:
                    max_idx = n.index
        # Save sparse vectors to dense vectors
        support_vectors = np.zeros(shape=(len(vectors), max_idx))
        for i, v in enumerate(vectors):
            for n in v.nodes:
                support_vectors[i][n.index - 1] = n.value
        support_vectors = support_vectors.flatten()
    else:
        raise ValueError('Unsupported support vector type: %s' % support_type)
    return len(vectors), support_vectors


def convert_svm_classifier(scope, operator, container):
    params = operator.raw_operator.supportVectorClassifier
    kernel_enum = {'linearKernel': 'LINEAR', 'polyKernel': 'POLY',
                   'rbfKernel': 'RBF', 'sigmoidKernel': 'SIGMOID', 'precomputedKernel': 'PRECOMPUTED'}
    kernel = params.kernel
    kernel_val = kernel.WhichOneof('kernel')
    svc_kernel = kernel_enum[kernel_val]

    if kernel_val == 'rbfKernel':
        svc_kernel_params = [kernel.rbfKernel.gamma, 0.0, 0.0]
    elif kernel_val == 'polyKernel':
        svc_kernel_params = [kernel.polyKernel.gamma,
                             kernel.polyKernel.coef0, kernel.polyKernel.degree]
    elif kernel_val == 'sigmoidKernel':
        svc_kernel_params = [kernel.sigmoidKernel.gamma,
                             kernel.sigmoidKernel.coef0, 0.0]
    elif kernel_val == 'linearKernel':
        svc_kernel_params = [0.0, 0.0, 0.0]

    prob_a = params.probA
    prob_b = params.probB
    support_vectors_per_class = params.numberOfSupportVectorsPerClass
    n_supports, svc_support_vectors = extract_support_vectors_as_dense_tensor(
        operator.raw_operator.supportVectorClassifier)
    chain_coef = list(itertools.chain.from_iterable([coef.alpha for coef in params.coefficients]))
    svc_coefficients = chain_coef
    svc_rho = [-x for x in params.rho]

    op_type = 'SVMClassifier'
    op_name = scope.get_unique_operator_name(op_type)
    attrs = {'name': op_name}
    attrs['kernel_type'] = svc_kernel
    attrs['kernel_params'] = svc_kernel_params
    if prob_a:
        attrs['prob_a'] = prob_a
    if prob_b:
        attrs['prob_b'] = prob_b
    attrs['vectors_per_class'] = support_vectors_per_class
    attrs['support_vectors'] = svc_support_vectors
    attrs['coefficients'] = svc_coefficients
    attrs['rho'] = svc_rho
    zipmap_attrs = {'name': scope.get_unique_operator_name('ZipMap')}
    svc_classes = params.WhichOneof('ClassLabels')
    if svc_classes == 'int64ClassLabels':
        class_labels = list(int(i) for i in params.int64ClassLabels.vector)
        attrs['classlabels_ints'] = class_labels
        zipmap_attrs['classlabels_int64s'] = class_labels
    elif svc_classes == 'stringClassLabels':
        class_labels = list(str(s).encode('utf-8') for s in params.stringClassLabels.vector)
        attrs['classlabels_strings'] = class_labels
        zipmap_attrs['classlabels_strings'] = class_labels
    else:
        raise ValueError('Unknown class label type')

    # For classifiers, due to the different representation of classes' probabilities, we need to add some
    # operators for type conversion. It turns out that we have the following topology.
    # input features ---> SupportVectorClassifier ---> label (must present)
    #                               |
    #                               '--> probability tensor ---> ZipMap ---> probability map (optional)

    raw_model = operator.raw_operator
    # Find label name and probability name
    proba_output_name = None
    for variable in operator.outputs:
        if raw_model.description.predictedFeatureName == variable.raw_name:
            label_output_name = variable.full_name
        if raw_model.description.predictedProbabilitiesName != '' and \
                raw_model.description.predictedProbabilitiesName == variable.raw_name:
            proba_output_name = variable.full_name

    inputs = [variable.full_name for variable in operator.inputs]
    proba_tensor_name = scope.get_unique_variable_name('ProbabilityTensor')

    if proba_output_name is not None:
        # Add support vector classifier in terms of ONNX node with probability output
        container.add_node(op_type, inputs, [label_output_name, proba_tensor_name], op_domain='ai.onnx.ml', **attrs)

        # Add ZipMap to convert probability tensor into probability map
        container.add_node('ZipMap', [proba_tensor_name], [proba_output_name],
                           op_domain='ai.onnx.ml', **zipmap_attrs)
    else:
        # Add support vector classifier in terms of ONNX node
        container.add_node(op_type, inputs, [label_output_name, proba_tensor_name], op_domain='ai.onnx.ml', **attrs)


register_converter('supportVectorClassifier', convert_svm_classifier)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import packaging.version as pv
import numpy
import onnx
from pyspark.ml.feature import StopWordsRemover
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import StringTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlStopWordsRemover(SparkMlTestCase):

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    @unittest.skipIf(pv.Version(onnx.__version__) <= pv.Version('1.5'),
                     'Need Greater Opset 10')
    def test_stop_words_remover2(self):
        data = self.spark.createDataFrame([(["a", "b", "c"],)], ["text"])
        model = StopWordsRemover(inputCol="text", outputCol="words", stopWords=["b"])

        model_onnx = convert_sparkml(model, 'Sparkml StopWordsRemover',
                                     [('text', StringTensorType([None]))], target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        expected = numpy.array(predicted.toPandas().words.values[0])
        data_np = numpy.array(data.toPandas().text.values[0])
        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlStopWordsRemover")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['words'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import sys
import inspect
import unittest
import packaging.version as pv
import onnx
import pandas
import numpy
from pyspark.ml.linalg import Vectors, SparseVector, VectorUDT
from pyspark.ml.regression import DecisionTreeRegressor
from pyspark.ml import Pipeline
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase
from pyspark.ml.feature import VectorIndexer


TARGET_OPSET = min(15, min(DEFAULT_OPSET_NUMBER, onnx_opset_version()))


class TestSparkmDecisionTreeRegressor(SparkMlTestCase):

    @unittest.skipIf(sys.platform == 'win32',
                     reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    @unittest.skipIf(pv.Version(onnx.__version__) <= pv.Version('1.3'), 'Need Greater Opset 9')
    def test_decision_tree_regressor_pipeline(self):
        import os
        this_script_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        input_path = os.path.join(this_script_dir, "data", "sample_libsvm_data.txt")
        original_data = self.spark.read.format("libsvm").load(input_path)

        feature_count = 5
        self.spark.udf.register("truncateFeatures",
                                lambda x: SparseVector(feature_count, range(0,feature_count), x.toArray()[125:130]),
                                VectorUDT())
        data = original_data.selectExpr("label", "truncateFeatures(features) as features")

        featureIndexer = \
            VectorIndexer(inputCol="features", outputCol="indexedFeatures", maxCategories=4, handleInvalid='error')
        (trainingData, testData) = data.randomSplit([0.7, 0.3])
        dt = DecisionTreeRegressor(featuresCol="indexedFeatures")
        pipeline = Pipeline(stages=[featureIndexer, dt])
        model = pipeline.fit(trainingData)
        model_onnx = convert_sparkml(model, 'Sparkml Decision Tree Regressor Pipeline', [
            ('features', FloatTensorType([None, feature_count]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(testData)
        data_np = testData.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32)
        ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlDecisionTreeRegressorPipeline")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.platform == 'win32',
                     reason="UnsatisfiedLinkError")
    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_decision_tree_regressor(self):
        features = [[0, 1], [1, 1], [2, 0]]
        features = numpy.array(features, dtype=numpy.float32)
        labels = [100, -10, 50]
        dd = [(labels[i], Vectors.dense(features[i])) for i in range(len(labels))]
        data = self.spark.createDataFrame(self.spark.sparkContext.parallelize(dd), schema=["label", "features"])
        dt = DecisionTreeRegressor(labelCol="label", featuresCol="features")
        model = dt.fit(data)
        feature_count = data.select('features').first()[0].size
        model_onnx = convert_sparkml(model, 'Sparkml Decision Tree Regressor', [
            ('features', FloatTensorType([None, feature_count]))
        ], spark_session=self.spark, target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        predicted = model.transform(data)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32)
        ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlDecisionTreeRegressor")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.data_types import FloatTensorType, Int64TensorType, FloatType, Int64Type
from ...common.utils import check_input_and_output_types


def calculate_traditional_regressor_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C] ---> [N, C']

    The number C' is the length of prediction vector. It can be a scalar (C'=1) or a vector (C'>1)
    '''
    check_input_and_output_types(operator, good_input_types=[FloatTensorType, Int64TensorType, FloatType, Int64Type])

    if any(len(variable.type.shape) != 2 for variable in operator.inputs):
        raise RuntimeError('Input(s) must be 2-D tensor(s)')

    model_type = operator.raw_operator.WhichOneof('Type')
    if model_type == 'glmRegressor':
        glm = operator.raw_operator.glmRegressor
        C = len(glm.weights)
    elif model_type == 'treeEnsembleRegressor':
        tree = operator.raw_operator.treeEnsembleRegressor.treeEnsemble
        C = len(tree.basePredictionValue)
    elif model_type == 'supportVectorRegressor':
        C = 1
    else:
        raise ValueError('Model should be one of linear model, tree-based model, and support vector machine')

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType([N, C], doc_string=operator.outputs[0].type.doc_string)

register_shape_calculator('glmRegressor', calculate_traditional_regressor_output_shapes)
register_shape_calculator('supportVectorRegressor', calculate_traditional_regressor_output_shapes)
register_shape_calculator('treeEnsembleRegressor', calculate_traditional_regressor_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ....proto import onnx_proto
from ...common._registration import register_converter


def convert_array_feature_extractor(scope, operator, container):
    op_type = 'ArrayFeatureExtractor'
    attrs = {'name': operator.full_name}

    target_indexes = operator.raw_operator.arrayFeatureExtractor.extractIndex
    index_buffer_name = scope.get_unique_variable_name('target_indexes')
    container.add_initializer(index_buffer_name, onnx_proto.TensorProto.INT64, [len(target_indexes)], target_indexes)

    inputs = [operator.inputs[0].full_name, index_buffer_name]
    outputs = [operator.outputs[0].full_name]

    container.add_node(op_type, inputs, outputs, op_domain='ai.onnx.ml', **attrs)


register_converter('arrayFeatureExtractor', convert_array_feature_extractor)



# SPDX-License-Identifier: Apache-2.0

from .....proto import helper
from .....proto import onnx_proto
from ....common._registration import register_converter
from ....common._apply_operation import apply_constant

def convert_load_constant(scope, operator, container):
    params = operator.raw_operator.loadConstant
    constant_name = scope.get_unique_variable_name('constant')
    constant = helper.make_tensor(constant_name, onnx_proto.TensorProto.FLOAT,
                                  params.shape, params.data.floatValue)

    apply_constant(scope, operator.output_full_names, container,
                   operator_name=operator.full_name, value=constant)

register_converter('loadConstant', convert_load_constant)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_concat, apply_argmax
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from .linear_classifier import convert_sparkml_linear_classifier
from ..ops_names import get_sparkml_operator_name
from ...common._registration import register_converter, register_shape_calculator


def convert_one_vs_rest(scope, operator, container):
    op = operator.raw_operator
    classifier_output_names = []
    # initializer needed to extract the 2nd value of probability array
    index_tensor = scope.get_unique_variable_name('index_tensor')
    container.add_initializer(index_tensor, onnx_proto.TensorProto.INT64, [1], [1])
    # OneVsRest could have different classifiers
    # all must have at least Probability values for us to do the argmax
    for sub_model in op.models:
        classifier_op = scope.declare_local_operator(get_sparkml_operator_name(type(sub_model)), sub_model)
        classifier_op.raw_params = operator.raw_params
        classifier_op.inputs = operator.inputs
        classifier_prediction_output = scope.declare_local_variable('classifier_prediction', FloatTensorType())
        classifier_probability_output = scope.declare_local_variable('classifier_probability', FloatTensorType())
        classifier_op.outputs.append(classifier_prediction_output)
        classifier_op.outputs.append(classifier_probability_output)
        convert_sparkml_linear_classifier(scope, classifier_op, container)
        classifier_op.is_evaluated = True
        single_feature_tensor = scope.get_unique_variable_name('single_feature_tensor')
        container.add_node('ArrayFeatureExtractor', [classifier_probability_output.full_name, index_tensor],
                           single_feature_tensor,
                           op_domain='ai.onnx.ml',
                           name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
        classifier_output_names.append(single_feature_tensor)

    concatenated_probabilities = scope.get_unique_variable_name('concatenated_predictions_tensor')
    apply_concat(scope, classifier_output_names, concatenated_probabilities, container, axis=1)
    # to get Prediction from probability
    apply_argmax(scope, concatenated_probabilities, operator.outputs[0].full_name, container,
                 axis=1, keepdims=1)


register_converter('pyspark.ml.classification.OneVsRestModel', convert_one_vs_rest)


def calculate_one_vs_rest_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType, Int64TensorType])
    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError('Input must be a [N, C]-tensor')

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = Int64TensorType(shape=[N])


register_shape_calculator('pyspark.ml.classification.OneVsRestModel',
                          calculate_one_vs_rest_output_shapes)
